==> kotest-framework-engine-jvm-5.5.1-sources.jar: jvmMain/io/kotest/engine/spec/createSpecExecutorDelegate.kt <==
package io.kotest.engine.spec

import io.kotest.common.ExperimentalKotest
import io.kotest.core.concurrency.CoroutineDispatcherFactory
import io.kotest.core.config.ProjectConfiguration
import io.kotest.core.spec.IsolationMode
import io.kotest.core.spec.Spec
import io.kotest.core.test.TestCase
import io.kotest.core.test.TestResult
import io.kotest.engine.concurrency.isIsolate
import io.kotest.engine.listener.TestEngineListener
import io.kotest.engine.spec.runners.InstancePerLeafSpecRunner
import io.kotest.engine.spec.runners.InstancePerTestSpecRunner
import io.kotest.engine.spec.runners.SingleInstanceSpecRunner
import io.kotest.engine.test.scheduler.ConcurrentTestScheduler
import io.kotest.engine.test.scheduler.SequentialTestScheduler
import io.kotest.mpp.Logger
import io.kotest.mpp.bestName
import kotlin.math.max

@ExperimentalKotest
internal actual fun createSpecExecutorDelegate(
   listener: TestEngineListener,
   defaultCoroutineDispatcherFactory: CoroutineDispatcherFactory,
   configuration: ProjectConfiguration,
): SpecExecutorDelegate = JvmSpecExecutorDelegate(listener, defaultCoroutineDispatcherFactory, configuration)

@ExperimentalKotest
class JvmSpecExecutorDelegate(
   private val listener: TestEngineListener,
   private val dispatcherFactory: CoroutineDispatcherFactory,
   private val configuration: ProjectConfiguration,
) : SpecExecutorDelegate {

   private val logger = Logger(JvmSpecExecutorDelegate::class)

   private fun Spec.resolvedIsolationMode() =
      this.isolationMode() ?: this.isolationMode ?: configuration.isolationMode

   override suspend fun execute(spec: Spec): Map<TestCase, TestResult> {

      val scheduler = when (val concurrentTests = spec.resolvedConcurrentTests(configuration.concurrentTests)) {
         ProjectConfiguration.Sequential -> SequentialTestScheduler
         else -> ConcurrentTestScheduler(max(1, concurrentTests))
      }

      val isolation = spec.resolvedIsolationMode()
      logger.log { Pair(spec::class.bestName(), "isolation=$isolation") }

      val runner = when (isolation) {
         IsolationMode.SingleInstance -> SingleInstanceSpecRunner(
            listener,
            scheduler,
            dispatcherFactory,
            configuration
         )
         IsolationMode.InstancePerTest -> InstancePerTestSpecRunner(
            listener,
            scheduler,
            dispatcherFactory,
            configuration
         )
         IsolationMode.InstancePerLeaf -> InstancePerLeafSpecRunner(
            listener,
            scheduler,
            dispatcherFactory,
            configuration
         )
      }

      return runner.execute(spec).getOrThrow()
   }
}

/**
 * Returns the concurrent tests count to use for tests in this spec.
 *
 * If threads is specified on the spec, then that will implicitly raise the concurrentTests
 * count to the same value if concurrentTests is not specified.
 *
 * Note that if this spec is annotated with @Isolate then the value
 * will be 1 regardless of the config setting.
 *
 * spec.concurrency ?: configuration.concurrentTests
 */
@OptIn(ExperimentalKotest::class)
internal fun Spec.resolvedConcurrentTests(defaultConcurrentTests: Int): Int {
   val fromSpecConcurrency = this.concurrency ?: this.concurrency()
   return when {
      this::class.isIsolate() -> ProjectConfiguration.Sequential
      fromSpecConcurrency != null -> max(1, fromSpecConcurrency)
      else -> defaultConcurrentTests
   }
}

==> kotest-framework-engine-jvm-5.6.2-sources.jar: jvmMain/io/kotest/engine/spec/createSpecExecutorDelegate.kt <==
package io.kotest.engine.spec

import io.kotest.common.ExperimentalKotest
import io.kotest.core.concurrency.CoroutineDispatcherFactory
import io.kotest.core.config.ProjectConfiguration
import io.kotest.core.spec.IsolationMode
import io.kotest.core.spec.Spec
import io.kotest.core.test.TestCase
import io.kotest.core.test.TestResult
import io.kotest.engine.concurrency.isIsolate
import io.kotest.engine.listener.TestEngineListener
import io.kotest.engine.spec.runners.InstancePerLeafSpecRunner
import io.kotest.engine.spec.runners.InstancePerTestSpecRunner
import io.kotest.engine.spec.runners.SingleInstanceSpecRunner
import io.kotest.engine.test.scheduler.ConcurrentTestScheduler
import io.kotest.engine.test.scheduler.SequentialTestScheduler
import io.kotest.mpp.Logger
import io.kotest.mpp.bestName
import kotlin.math.max

@ExperimentalKotest
internal actual fun createSpecExecutorDelegate(
   listener: TestEngineListener,
   defaultCoroutineDispatcherFactory: CoroutineDispatcherFactory,
   configuration: ProjectConfiguration,
): SpecExecutorDelegate = JvmSpecExecutorDelegate(listener, defaultCoroutineDispatcherFactory, configuration)

@ExperimentalKotest
class JvmSpecExecutorDelegate(
   private val listener: TestEngineListener,
   private val dispatcherFactory: CoroutineDispatcherFactory,
   private val configuration: ProjectConfiguration,
) : SpecExecutorDelegate {

   private val logger = Logger(JvmSpecExecutorDelegate::class)

   private fun Spec.resolvedIsolationMode() =
      this.isolationMode() ?: this.isolationMode ?: configuration.isolationMode

   override suspend fun execute(spec: Spec): Map<TestCase, TestResult> {

      val scheduler = when (val concurrentTests = spec.resolvedConcurrentTests(configuration.concurrentTests)) {
         ProjectConfiguration.Sequential -> SequentialTestScheduler
         else -> ConcurrentTestScheduler(max(1, concurrentTests))
      }

      val isolation = spec.resolvedIsolationMode()
      logger.log { Pair(spec::class.bestName(), "isolation=$isolation") }

      val runner = when (isolation) {
         IsolationMode.SingleInstance -> SingleInstanceSpecRunner(
            listener,
            scheduler,
            dispatcherFactory,
            configuration
         )
         IsolationMode.InstancePerTest -> InstancePerTestSpecRunner(
            listener,
            scheduler,
            dispatcherFactory,
            configuration
         )
         IsolationMode.InstancePerLeaf -> InstancePerLeafSpecRunner(
            listener,
            scheduler,
            dispatcherFactory,
            configuration
         )
      }

      return runner.execute(spec).getOrThrow()
   }
}

/**
 * Returns the concurrent tests count to use for tests in this spec.
 *
 * If threads is specified on the spec, then that will implicitly raise the
 * [concurrentTests][io.kotest.core.config.AbstractProjectConfig.concurrentTests]
 * count to the same value if
 * [concurrentTests][io.kotest.core.config.AbstractProjectConfig.concurrentTests]
 * is not specified.
 *
 * Note that if this spec is annotated with [@Isolate][io.kotest.core.annotation.Isolate] then
 * the value will be 1 regardless of the config setting.
 *
 * ```
 * spec.concurrency ?: configuration.concurrentTests
 * ```
 */
@OptIn(ExperimentalKotest::class)
internal fun Spec.resolvedConcurrentTests(defaultConcurrentTests: Int): Int {
   val fromSpecConcurrency = this.concurrency ?: this.concurrency()
   return when {
      this::class.isIsolate() -> ProjectConfiguration.Sequential
      fromSpecConcurrency != null -> max(1, fromSpecConcurrency)
      else -> defaultConcurrentTests
   }
}
